#ifndef __MBUFF_H
#define __MBUFF_H

#define RTL_SHM_MISC_MINOR 254

#define MBUFF_NAME_LEN 32
#define MBUFF_MAX_MMAPS 5

/*
   All ioctl()s are called with name filled in with the appropriate
   name for the mbuff to be referenced.  Calls to any ioctl() makes
   that mbuff "active", i.e., read(), write(), and mmap() use that
   mbuff.  I didn't do this yet.

   ioctl()s:

   ALLOCATE:
      Call with size=0 to just find out if the area exists; no
      mbuff will be allocated.  Otherwise, allocate an mbuff with
      that size.
   DEALLOCATE:
      Decrease reference count for an mbuff.

   issues:
      - using this method, it is *really* easy to get dangling
        mbuffs, i.e., mbuffs that nobody owns.  When you close
	/dev/mbuff, it would be a good idea to decrease the ref
	count of the active mbuff.
 */
#define IOCTL_MBUFF_INFO 0
#define IOCTL_MBUFF_ALLOCATE 1
#define IOCTL_MBUFF_DEALLOCATE 2
#define IOCTL_MBUFF_SELECT 3
#define IOCTL_MBUFF_LAST IOCTL_MBUFF_SELECT

struct mbuff_request_struct{
	unsigned int flags;

	char name[MBUFF_NAME_LEN+1];

	size_t size;

	unsigned int reserved[4];
};

#ifdef __KERNEL__
struct mbuff{
	struct mbuff *next;
	struct mbuff *prev;

	char name[MBUFF_NAME_LEN+1];

	struct vm_area_struct *(vm_area[MBUFF_MAX_MMAPS]);
	struct file *file;

	unsigned char *buf;
	unsigned long size;
	int count;
	int open_cnt;	/* #times opened */
	int open_mode;
};

extern int shm_allocate(const char *name,unsigned int size, void **shm);
extern int shm_deallocate(void * shm);
extern char mbuff_default_name[];
extern int mbuff_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg);
#ifdef LINUX_V22
extern int mbuff_mmap(struct file *file, struct vm_area_struct *vma);
#else
extern int mbuff_mmap(struct inode *inode, struct file *file, 
	struct vm_area_struct *vma);
#endif
extern int mbuff_open_with_name( struct inode *inode, struct file *file,
	const char * name);


#endif
#endif
