/*
 * Main system for RTL.
 * 
 * Copyright (C) 1999 Cort Dougan <cort@fsmlabs.com>
 * Copyright (C) 1999 Victor Yodaiken <yodaiken@fsmlabs.com>
 */

#ifndef __RTL_CORE_H__
#define __RTL_CORE_H__

#include <linux/threads.h>
#include <linux/sched.h>
#include <rtl_printf.h>
#include "arch/constants.h"

/* rtl_irqstate_t needs to go into arch */
typedef unsigned long rtl_irqstate_t;

intercept_t  rtl_intercept(MACHDEPREGS regs);
void rtl_soft_sti(void);
void rtl_soft_cli(void);
void rtl_soft_save_flags(unsigned long *x);
void rtl_soft_restore_flags(unsigned long x);
void rtl_global_pend_irq(int);
void rtl_virt_disable(unsigned int);
void rtl_virt_enable(unsigned int);
void rtl_process_pending(void);

#ifdef __LOCAL_IRQS__
void dispatch_local_linux_irq(struct pt_regs *);
void rtl_local_pend_vec(int vector,int cpu_id);
int rtl_request_local_irq(int i, unsigned int (*handler)(struct pt_regs *r), unsigned int cpu);
void soft_dispatch_local(unsigned int);
int rtl_free_local_irq(int i, unsigned int cpu);
#endif
int arch_takeover(void);
void arch_giveup(void);
void rtl_hard_enable_irq(unsigned int);
int rtl_request_global_irq(unsigned int irq, unsigned int (*handler)(unsigned int, struct pt_regs *));
int rtl_free_global_irq(unsigned int irq );
int rtl_get_soft_irq (void (*handler) (int, void *, struct pt_regs *),
		      const char *devname);

#define rtl_request_irq rtl_request_global_irq
#define rtl_free_irq rtl_free_global_irq

void rtl_make_rt_system_active(void);
void rtl_make_rt_system_idle(void);
unsigned int rtl_rt_system_is_idle(void);

#define rtl_free_soft_irq(irq) free_irq(irq, 0)

struct rtl_local {
	__u32    flags;
#ifdef __LOCAL_IRQS__
	__u32    pending;
	__u32    rtirq;
#endif
};

#define rtl_getcpuid() hard_smp_processor_id()
#define rtl_num_cpus() (smp_num_cpus)

extern int rtl_cpu_exists (int cpu);

extern struct rtl_local rtl_local[NR_CPUS];
void conprn(const unsigned int hexnum);
void conpr(const char *s);
#endif
