/*
 * Alpha-specific parts of the main system for RTL.
 * 
 * Copyright (C) 2000 FSM Labs (http://www.fsmlabs.com/)
 *  Written by Cort Dougan <cort@fsmlabs.com>
 */
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/irq.h>
#include <asm/irq.h>
#include <asm/machvec.h>

#include "arch.h"
#include <rtl_core.h>
#include <rtl_sync.h>

extern unsigned long linux_soft_sti;

irq_desc_t rtl_hard_irq_desc[NR_IRQS];
void (*hard_do_IRQ)(unsigned long, struct pt_regs *);
void rtl_local_intercept(struct pt_regs *regs);
int last_irq = -1;

void alpha_ack(unsigned int irq)
{
}

void alpha_end(unsigned int irq)
{
	rtl_virt_enable(irq);
}

unsigned int alpha_startup(unsigned int irq)
{
	rtl_virt_enable(irq);
	return 0;
}

hw_irq_controller rtl_fake_pic =
{ " RTL PIC  ", alpha_startup, rtl_virt_disable, rtl_virt_enable,
  rtl_virt_disable, alpha_ack, alpha_end };

unsigned long alpha_soft_rdps(void)
{
	unsigned long x;
	rtl_soft_save_flags(&x);
	return (rdps()&~7) | (x&7);
}

unsigned long alpha_soft_swpipl(unsigned long ipl)
{
	unsigned long prev = alpha_soft_rdps()&7;
	extern unsigned long last_cli;
	ipl &= 7;
	
	/* special case since IPL_MIN can be a variable */
	if ( ipl == IPL_MIN )
	{
		rtl_soft_sti();
		return prev;
	}
	switch ( ipl )
	{
	case IPL_MAX:
		rtl_soft_cli();
		last_cli = (ulong)__builtin_return_address(0);
		break;
	case 0:
		rtl_soft_sti();
		break;
	default:
		rtl_printf("alpha_soft_setipl: unknown IPL %x\n", ipl );
		break;
	}
	return prev;
}

void alpha_intercept(unsigned long irq, struct pt_regs *regs)
{
	if ( last_irq != -1 )
		printk("alpha_intercept: last_irq set\n");
	last_irq = irq;
	rtl_intercept(regs);
}

int arch_takeover(void)
{
	int i;
	void rtl_soft_sti_no_emulation(void);

	__cli();
	memcpy(rtl_hard_irq_desc, irq_desc, sizeof(irq_desc_t)*NR_IRQS);
	__rdps = alpha_soft_rdps;
	__swpipl = alpha_soft_swpipl;

	for ( i = 0 ; i < NR_IRQS; i++ )
		if ( irq_desc[i].handler )
			irq_desc[i].handler = &rtl_fake_pic;

	hard_do_IRQ = handle_irq;
	handle_irq = alpha_intercept;
	linux_soft_sti = (ulong)rtl_local;
	rtl_hard_sti();
	return 0;
}

void arch_giveup(void)
{
	rtl_hard_cli();
	__rdps = rdps;
	__swpipl = swpipl;
	memcpy(irq_desc, rtl_hard_irq_desc, sizeof(irq_desc_t)*NR_IRQS);
	linux_soft_sti = (ulong)0;
	handle_irq = hard_do_IRQ;
	__sti();
}

void dispatch_rtl_local_irq(int irq)
{
}

int rtl_free_local_irq(int i, unsigned int cpu)
{
	return 0;
}

int rtl_request_local_irq(int i, unsigned int (*handler)(struct pt_regs *r),
			  unsigned int cpu)
{
	return 0;
}
