/*
 * RTL ALPHA-specific header.
 *
 * Copyright (C) 1999 FSM Labs (http://www.fsmlabs.com/)
 *  Written by Cort Dougan <cort@fsmlabs.com>
 */
#ifndef _ALPHA_ARCH_H
#define _ALPHA_ARCH_H
#include <linux/irq.h>
#include <linux/config.h>
#include <arch/constants.h>
#include <rtl_core.h>

#ifndef CONFIG_RTLINUX
#error CONFIG_RTLINUX is not set in the kernel source
#endif

extern void (*hard_do_IRQ)(unsigned long, struct pt_regs *);
/*
 * this is in arch/alpha/kernel/irq_impl.h so we have our own
 * prototype here -- Cort
 */
extern void (*handle_irq)(unsigned long, struct pt_regs *);

#define LOCAL_TIMER_VECTOR 0
#define RTL_RESCHEDULE_VECTOR 1 /* alpha doesn't use this */
#define soft_local_disabled_irq(irq) (0) /* cannot disable timer */
extern int last_irq;

/*
 * Global dispatch functions
 */
#define dispatch_rtl_handler(irq,regs)
#define dispatch_linux_irq(regs,irq) { hard_do_IRQ(irq,regs); rtl_virt_enable(irq); }
extern __inline__ void soft_dispatch_global(unsigned int irq)
{
	struct pt_regs regs;
	hard_do_IRQ(irq,&regs);
	rtl_virt_enable(irq);
}

/*
 * Local irq functions
 */ 
#define dispatch_rtl_local_handler(pnd,regs) timer_handler(regs)
#define dispatch_local_linux_irq(regs) timer_interrupt(regs)
extern __inline__ void soft_dispatch_local(unsigned int irq)
{
}

#define ARCH_DEFINED_ENABLE (IPL_MIN)
#define ARCH_DEFINED_DISABLE (IPL_MAX)

extern irq_desc_t rtl_hard_irq_desc[NR_IRQS];

#define rtl_irq_controller_enable(irq) \
({ if (rtl_hard_irq_desc[irq].handler && rtl_hard_irq_desc[irq].handler->enable) \
	rtl_hard_irq_desc[irq].handler->enable(irq); })
#define rtl_irq_controller_disable(irq) \
({ if (rtl_hard_irq_desc[irq].handler && rtl_hard_irq_desc[irq].handler->disable) \
	rtl_hard_irq_desc[irq].handler->disable(irq); })
#define rtl_irq_controller_get_irq(regs) \
({ int i = last_irq; last_irq = -1; i; })
#define rtl_irq_controller_postirq(regs,irq)
#define rtl_irq_controller_ack(irq) \
({ if (rtl_hard_irq_desc[irq].handler && rtl_hard_irq_desc[irq].handler->ack) \
	rtl_hard_irq_desc[irq].handler->ack(irq); })
#define rtl_local_irq_controller_ack() { }

#endif /* _ALPHA_ARCH_H */
