/*
 * RTL PPC-specific header.
 *
 * Copyright (C) 1999 FSM Labs (http://www.fsmlabs.com/)
 *  Written by Cort Dougan <cort@fsmlabs.com>
 */
#ifndef _PPC_ARCH_H
#define _PPC_ARCH_H
#include <linux/irq.h>
#include <arch/constants.h>
#include <rtl_core.h>

#define LOCAL_TIMER_VECTOR 0
#define RTL_RESCHEDULE_VECTOR 1 /* ppc doesn't use this */
#define soft_local_disabled_irq(irq) (0) /* cannot disable timer */

extern unsigned int (*timer_handler)(struct pt_regs *r);

/*
 * Global dispatch functions
 */
void ppc_irq_dispatch_handler(struct pt_regs *regs, int irq);
#define dispatch_rtl_handler(irq,regs)
#define dispatch_linux_irq(regs,irq) ppc_irq_dispatch_handler(regs,irq)
extern __inline__ void soft_dispatch_global(unsigned int irq)
{
	struct pt_regs fake_regs;
	ppc_irq_dispatch_handler(&fake_regs, irq);
}

/*
 * Local irq functions
 */ 
#define dispatch_rtl_local_handler(pnd,regs) timer_handler(regs)
#define dispatch_local_linux_irq(regs) timer_interrupt(regs)
extern __inline__ void soft_dispatch_local(unsigned int irq)
{
	struct pt_regs regs;
	dispatch_local_linux_irq(&regs);
}

#define ARCH_DEFINED_ENABLE (MSR_EE)
#define ARCH_DEFINED_DISABLE 0

extern irq_desc_t rtl_hard_irq_desc[NR_IRQS];
extern struct int_control_struct rtl_hard_int_control;
extern int (*hard_get_irq)(struct pt_regs *);
extern void (*hard_post_irq)(struct pt_regs *,int);

#define rtl_irq_controller_enable(irq) \
({ if (rtl_hard_irq_desc[irq].handler && rtl_hard_irq_desc[irq].handler->enable) \
	rtl_hard_irq_desc[irq].handler->enable(irq); })
#define rtl_irq_controller_disable(irq) \
({ if (rtl_hard_irq_desc[irq].handler && rtl_hard_irq_desc[irq].handler->disable) \
	rtl_hard_irq_desc[irq].handler->disable(irq); })
#define rtl_irq_controller_get_irq(regs) hard_get_irq(regs)
#define rtl_irq_controller_postirq(regs,irq) \
({ if ( hard_post_irq )	hard_post_irq(regs,irq); })
#define rtl_irq_controller_ack(irq) \
({ if (rtl_hard_irq_desc[irq].handler && rtl_hard_irq_desc[irq].handler->ack) \
	rtl_hard_irq_desc[irq].handler->ack(irq); })
#define rtl_local_irq_controller_ack() { }

#endif /* _PPC_ARCH_H */
