/*
 * RTL core features.
 *
 * Copyright (C) 1999 FSM Labs (http://www.fsmlabs.com/)
 *  Written by Cort Dougan <cort@fsmlabs.com> and
 *  Victor Yodaiken <yodaiken@fsmlabs.com>
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/timex.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/console.h>
#include <linux/irq.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/segment.h>

#include <arch/constants.h>
#include <rtl_conf.h>
#include <rtl_printf.h>
#include <rtl_core.h>
#include <rtl_sync.h>
#include <rtl.h>
#include <linux/irq.h>

void rtl_hard_disable_irq(unsigned int ix);
static inline void debug_test_enabled(char *s);
unsigned long last_cli;

void conpr(const char *s)
{
	long flags;
	static spinlock_t rtl_conpr_lock = SPIN_LOCK_UNLOCKED;
	struct console *c;
	int len = strlen(s);

	rtl_hard_savef_and_cli(flags);
	spin_lock(&rtl_conpr_lock);

	c = console_drivers;
	while(c) {
		if ((c->flags & CON_ENABLED) && c->write)
			c->write(c, s, len);
		c = c->next;
	}
	spin_unlock(&rtl_conpr_lock);
	rtl_hard_restore_flags(flags);
}

void conprn(const unsigned int hexnum)
{
	int i;
	unsigned int d;
	unsigned int n = hexnum;
    	char s[10];
	s[9] = 0;
	s[8] = ' ';
	for (i=7; i>=0; i--) {
		d = n % 16;
		if (d < 10) {
			d += '0';
		} else {
			d += 'a' - 10;
		}
		s[i] = d; 
		n = n / 16;
	}
    	conpr(s);
}

irq_desc_t *linux_irq_desc; /* initialized in arch dependent code*/
/* assuming 255 global irqs and 31 max local vectors 
   On the x86 we only have local irqs when we are smp.
   But in PowerPC and other we may have local irqs from
   on chip timers and other advanced technologies
   */

/* bit positions for flags, constants, and macros for global structure */
#define IRQ_NOT_VALID -1
#define IRQ_ARRAY_SIZE ((256/8)/sizeof(ulong)) /*256 global irqs */
#define IRQ_ZINIT {0}
#if BITS_PER_LONG == 32
#define IRQ_NZINIT {~0x0,~0x0,~0x0,~0x0,~0x0,~0x0,~0x0,~0x0}
#else
#define IRQ_NZINIT {~0x0,~0x0,~0x0,~0x0}
#endif

/* fix if IRQ_ARRAY_SIZE !=4  */
#if BITS_PER_LONG == 32
#define irq_toi(x)  ((x>>5)&7)
#define irq_top(x)  ((x)& 0x1fUL)
#else
#define irq_toi(x)  ((x>>6)&15)
#define irq_top(x)  ((x)& 0x3fUL)
#endif
#define pi_toirq(p,i) ( (p) + (i*BITS_PER_LONG))

#define G_PEND(f) set_bit(irq_top(f),&rtl_global.pending[irq_toi(f)])
#define G_UNPEND(f) clear_bit(irq_top(f),&rtl_global.pending[irq_toi(f)])
#define G_ISPEND(f) test_bit(irq_top(f),&rtl_global.pending[irq_toi(f)])
/* clear and set global enabled irq bits */
#define G_ENABLED(f) set_bit(irq_top(f),&rtl_global.soft_enabled[irq_toi(f)])
#define G_DISABLE(f) clear_bit(irq_top(f),&rtl_global.soft_enabled[irq_toi(f)])
#define G_ISENABLED(f) test_bit(irq_top(f),&rtl_global.soft_enabled[irq_toi(f)])
/* clear and set real time handlers (RealTimeHandlers) */
#define G_SET_RTH(f) set_bit(irq_top(f),&rtl_global.rtirq[irq_toi(f)])
#define G_CLEAR_RTH(f) clear_bit(irq_top(f),&rtl_global.rtirq[irq_toi(f)])
#define G_TEST_RTH(f) test_bit(irq_top(f),&rtl_global.rtirq[irq_toi(f)])
#define G_TEST_AND_SET_RTH(f) test_and_set_bit(irq_top(f),&rtl_global.rtirq[irq_toi(f)])
#define G_TEST_AND_CLEAR_RTH(f) test_and_clear_bit(irq_top(f),&rtl_global.rtirq[irq_toi(f)])
/* global flags */
#define g_rtl_started 0
#define g_pend_since_sti 1
#define g_initializing 2
#define g_initialized 3
#define G_SET(f) set_bit(f,&rtl_global.flags)
#define G_CLEAR(f) clear_bit(f,&rtl_global.flags)
#define G_TEST(f) test_bit(f,&rtl_global.flags)
#define G_TEST_AND_SET(f) test_and_set_bit(f,&rtl_global.flags)
#define G_TEST_AND_CLEAR(f) test_and_clear_bit(f,&rtl_global.flags)

/* Bit positions of flags for local structure and macros
   for operating on them
 */
#define l_idle 0
#define l_ienable 1
#define l_pend_since_sti 2
#define l_busy 3
#define L_SET(f) set_bit(f,&rtl_local[cpu_id].flags)
#define L_CLEAR(f) clear_bit(f,&rtl_local[cpu_id].flags)
#define L_TEST(f) test_bit(f,&rtl_local[cpu_id].flags)
#define L_TEST_AND_SET(f) test_and_set_bit(f,&rtl_local[cpu_id].flags)
#define L_PEND(f) set_bit(f,&rtl_local[cpu_id].pending)
#define L_UNPEND(f) clear_bit(f,&rtl_local[cpu_id].pending)
#define L_ISPEND(f) test_bit(f,&rtl_local[cpu_id].pending)

#define L_SET_RTH(f) set_bit(f,&rtl_global.rtirq)
#define L_CLEAR_RTH(f) clear_bit(f,&rtl_local[cpu_id].rtirq)
#define L_TEST_RTH(f) test_bit(f,&rtl_local[cpu_id].rtirq)
#define L_TEST_AND_SET_RTH(f) test_and_set_bit(f,&rtl_local[cpu_id].rtirq)
#define L_TEST_AND_CLEAR_RTH(f) test_and_clear_bit(f,&rtl_local[cpu_id].rtirq)
void rtl_debug(void);

#ifdef  __SMP__
#undef rtl_spin_lock
#undef rtl_spin_unlock
inline static void rtl_spin_lock(spinlock_t *s)
{
	int count;
	for(count=0;count<100000000;count++)
	{
	     	if(!test_and_set_bit(0,&s->lock))
		       	break;
	       	if(count==100000000)
		{
			/*conpr("\nRTLBADSpin=");
			  conprn((unsigned int)&count);*/
		       	panic("badspin"); 
		}
	}
}

inline static void rtl_spin_unlock(spinlock_t *s)
{
	clear_bit(0,&s->lock);				
}
#define rtl_linux_spin_lock(x) {rtl_hard_cli(); rtl_spin_lock(x);}
#define rtl_linux_spin_unlock(x) {rtl_spin_unlock(x);rtl_hard_sti();}
/* TODO soft smp_processor_id doesn't work here???? */
#define DeclareAndInit(cpu_id)  unsigned int cpu_id = hard_smp_processor_id()
#define HardDeclareAndInit(cpu_id)  unsigned int cpu_id = hard_smp_processor_id()

#else
#define rtl_spin_lock(x)  
#define rtl_spin_unlock(x) 
#define rtl_linux_spin_lock(x) rtl_hard_cli(); 
#define rtl_linux_spin_unlock(x) rtl_hard_sti();
#define DeclareAndInit(cpu_id) const int cpu_id=0;
#define HardDeclareAndInit(cpu_id) const int cpu_id=0;

#endif __SMP__

/* The basic control data structures local and global*/
struct rtl_local rtl_local[NR_CPUS];

struct rtl_global{
	spinlock_t hard_irq_controller_lock;
	unsigned long flags;
	unsigned long pending[IRQ_ARRAY_SIZE];
	unsigned long soft_enabled[IRQ_ARRAY_SIZE];
	unsigned long rtirq[IRQ_ARRAY_SIZE];
};
struct rtl_global rtl_global ={ SPIN_LOCK_UNLOCKED,0,IRQ_ZINIT,IRQ_NZINIT,IRQ_ZINIT} ;

/* rtl interrupts */
struct rtl_global_handlers{
        unsigned int (*handler)(unsigned int irq, struct pt_regs *r);
}rtl_global_handlers[IRQ_MAX_COUNT];

#ifdef __LOCAL_IRQS__
#define LOCAL_PND_MAX 32
struct rtl_local_handlers{
        unsigned int (*handler)(struct pt_regs *r);
} rtl_local_handlers[LOCAL_PND_MAX];
void rtl_local_pend_vec(int vector,int cpu_id) {
	int i = VECTOR_TO_LOCAL_PND(vector);
	L_PEND(i);
	L_SET(l_pend_since_sti);
}
#endif

extern void * rtl_code[];
#include "arch/arch.h"

static inline void debug_test_enabled(char *s)
{
	unsigned long flags;

	rtl_hard_save_flags(flags);
	if ( (ARCH_DEFINED_ENABLE && !(flags & ARCH_DEFINED_ENABLE))
	     || (!ARCH_DEFINED_ENABLE && (flags & ARCH_DEFINED_DISABLE)) )
	{
		rtl_printf("%s: intrs hard disabled! called from %p\n",\
			   s, __builtin_return_address(0));
	}
}

/* rtl_intercept intercepts global interrupts */
#define RUN_LINUX_HANDLER(irq) (G_ISPEND(irq) && !L_TEST(l_busy)\
       	&& L_TEST(l_ienable) && G_ISENABLED(irq))
intercept_t rtl_intercept(MACHDEPREGS regs)
{
	int irq;
	HardDeclareAndInit(cpu_id);
	rtl_spin_lock(&rtl_global.hard_irq_controller_lock);
	if ((irq = rtl_irq_controller_get_irq(regs)) != -1)
	{
		rtl_irq_controller_ack(irq); /* may also mask, if needed */
		
		if(G_TEST_RTH(irq)){ /* this is a RT irq */
			/* if RT wants to share it pends */
			dispatch_rtl_handler(irq,&regs);
		} else {
			G_PEND(irq);
			G_SET(g_pend_since_sti);
		}
		if(RUN_LINUX_HANDLER(irq))
		{
			/* unpend so dispatch doesn't dispatch 2 times*/
			G_UNPEND(irq); 
			L_CLEAR(l_ienable); /* disable local soft interrupts */
			G_DISABLE(irq); /* disable this irq */
			rtl_irq_controller_postirq(MACHDEPREGS_PTR(regs),irq);
			rtl_spin_unlock(&rtl_global.hard_irq_controller_lock);
			rtl_hard_sti(); /* yikes ! */
			dispatch_linux_irq(MACHDEPREGS_PTR(regs),irq);
			rtl_soft_sti();
			RETURN_FROM_INTERRUPT_LINUX; /* goes via ret_from_intr */
		}
		rtl_irq_controller_postirq(MACHDEPREGS_PTR(regs),irq);
	}
	/* get here if irq==-1 or if otherwise can't run linux handler */
	rtl_spin_unlock(&rtl_global.hard_irq_controller_lock);
	RETURN_FROM_INTERRUPT;
}

#ifdef __LOCAL_IRQS__
static inline unsigned int get_lpended_irq(void)
{
	int i;
	DeclareAndInit(cpu_id); /* only called in Linux context */
	if(rtl_local[cpu_id].pending){
		i = ffz(~rtl_local[cpu_id].pending);
		clear_bit(i,&rtl_local[cpu_id].pending);
		i = LOCAL_PND_TO_VECTOR(i);
	}
	else i = IRQ_NOT_VALID;
/* 	conprn(i); */
/* 	conprn(rtl_local[cpu_id].pending); */
	return i ;
}

intercept_t rtl_local_intercept(MACHDEPREGS regs)
{
	int pnd;

	HardDeclareAndInit(cpu_id);
	/* no lock needed because we are already hard cli and only
	   use local per-cpu structures. The rtl_irq_controller
	   operations MUST vector to local only hardware or must
	   use spinlocks */
	pnd = MACHDEPREGS_TO_PND(regs);
	rtl_local_irq_controller_ack();
	if(L_TEST_RTH(pnd)){ /* this is a RT irq */
		dispatch_rtl_local_handler(pnd,MACHDEPREGS_PTR(regs));/* if RT wants to share it pends */
	}
	else{
		L_PEND(pnd);
/* 		conpr("p"); */
/* 		conprn(rtl_local[cpu_id].pending); */
		L_SET(l_pend_since_sti);
	}
	/* VY just removed test for soft local disabled. I don't
	   think we have soft local disabled */
	if (!L_TEST(l_ienable)) {
/* 		conpr("r"); */
	}
	if(!L_ISPEND(pnd) || L_TEST(l_busy) || !L_TEST(l_ienable) )
	{
		RETURN_FROM_LOCAL;
	}
	else
	{
/* 		conpr("u"); */
		L_UNPEND(pnd); /* yes it is stupid, see above */
		L_CLEAR(l_ienable); /* disable local soft interrupts */
		rtl_hard_sti(); /* yikes ! */
		dispatch_local_linux_irq(MACHDEPREGS_PTR(regs));
	}
	RETURN_FROM_LOCAL_LINUX;
}
#endif

/* tools for soft_sti */
static inline unsigned int get_gpended_irq(void)
{
	unsigned int i;
	rtl_linux_spin_lock(&rtl_global.hard_irq_controller_lock);
	/*obviously this loop should be optimized instead of 
	  searching bit by bit but costs are not much greater */

	for(i=0; i < IRQ_MAX_COUNT; i ++){
		if(G_ISPEND(i) && G_ISENABLED(i)){
			G_UNPEND(i); /* chance for a race, but
					we don't care. If other CPU disables
					while the irq is in transit
					outcome is not determined anyways */
			break;
		}
	}
	rtl_linux_spin_unlock(&rtl_global.hard_irq_controller_lock);
	return (i < IRQ_MAX_COUNT ? i: IRQ_NOT_VALID);
}

void rtl_soft_cli(void)
{
	DeclareAndInit(cpu_id);
	L_CLEAR(l_ienable);
	last_cli = (unsigned long)__builtin_return_address(0);
}

void rtl_soft_sti_no_emulation(void)
{
	DeclareAndInit(cpu_id);
	L_SET(l_ienable);
}

#define RTL_MAX_IRQ_DEPTH 25

void rtl_process_pending(void)
{
	int irq = 0;
	int last_irq = 0;
	int dispatch_count = 0;
	static int depth_count = 0;
	DeclareAndInit(cpu_id);
	
	L_CLEAR(l_ienable); /*disable soft interrupts !*/
       	do{
		irq = IRQ_NOT_VALID;
	       	G_CLEAR(g_pend_since_sti);
	       	L_CLEAR(l_pend_since_sti);
#ifdef __LOCAL_IRQS__
	       	if( (irq = get_lpended_irq()) != IRQ_NOT_VALID )
			soft_dispatch_local(irq);
#endif
	       	if( (irq = get_gpended_irq()) != IRQ_NOT_VALID )
		{
			++depth_count;
			++dispatch_count;
			last_irq = irq;
			soft_dispatch_global(irq);
			--depth_count;
		}
/* XXX:
 * This loop will not handle the case of 2 local irqs
 * and no global correctly.
 * -- Cort
 */
	}while(irq != IRQ_NOT_VALID || G_TEST(g_pend_since_sti) ||
		L_TEST(l_pend_since_sti));
}

void rtl_soft_sti(void)
{
	debug_test_enabled("rtl_soft_sti");
	rtl_process_pending();
	rtl_soft_sti_no_emulation();
}

void rtl_soft_save_flags(unsigned long *x)
{
	DeclareAndInit(cpu_id);
	*x = (L_TEST(l_ienable)? ARCH_DEFINED_ENABLE: ARCH_DEFINED_DISABLE);
}

void rtl_soft_restore_flags(unsigned long x)
{
	if(x == ARCH_DEFINED_ENABLE)rtl_soft_sti();
	else rtl_soft_cli();
}

void rtl_soft_local_irq_save(unsigned long *x)
{
	rtl_soft_save_flags(x);
	rtl_soft_cli();
}

void rtl_soft_local_irq_restore(unsigned long x)
{
	rtl_soft_restore_flags(x);
}

void rtl_virt_disable(unsigned int irq)
{
	G_DISABLE(irq);
}

void rtl_virt_enable(unsigned int irq)
{
	G_ENABLED(irq);
	if(!G_ISPEND(irq)){
		rtl_hard_enable_irq(irq);
	}
	else{
		HardDeclareAndInit(cpu_id);
		if( L_TEST(l_ienable))
			__sti(); /* emulate the bastard */
	}
}

/* these are exported so that they can be called by rt drivers */
void rtl_global_pend_irq(int ix) { G_PEND(ix);}

void rtl_hard_enable_irq(unsigned int ix)
{
	rtl_spin_lock(&rtl_global.hard_irq_controller_lock);
	rtl_irq_controller_enable(ix);
	rtl_spin_unlock(&rtl_global.hard_irq_controller_lock);
}

void rtl_hard_disable_irq(unsigned int ix)
{
	rtl_spin_lock(&rtl_global.hard_irq_controller_lock);
	rtl_irq_controller_disable(ix);
	rtl_spin_unlock(&rtl_global.hard_irq_controller_lock);
}

/* these are used by schedulers to make sure that Linux interrupts
   do not advance and delay RT tasks 
 Both  need to be called with irqs disabled */
void rtl_make_rt_system_active(void)
{
	HardDeclareAndInit(cpu_id);
	L_SET(l_busy);
}

void rtl_make_rt_system_idle(void)
{
	HardDeclareAndInit(cpu_id);
        L_CLEAR(l_busy);

}

unsigned int rtl_rt_system_is_idle(void)
{
	HardDeclareAndInit(cpu_id);
        return !L_TEST(l_busy);

}

/* requesting and freeing rt interrupts */
/* TODO resolve the smp synchronization problem here */
int rtl_request_global_irq(unsigned int irq, 
			   unsigned int (*handler)(unsigned int, struct pt_regs *))
{

	if (!G_TEST_RTH(irq)) {
		rtl_global_handlers[irq].handler =handler;
		G_SET_RTH(irq);
		mb();
		if(rtl_global_handlers[irq].handler == handler){
			rtl_hard_enable_irq (irq);
			return 0;
		}
	}
	return -EBUSY;
}

int rtl_free_global_irq(unsigned int irq )
{
	if (!G_TEST_AND_CLEAR_RTH(irq)) {
		return -EINVAL;
	}
	return 0;
	/* don't need to clear the handler, because it will never
	   be invoked -- see rtl_intercept. If we wanted to clear the handler
	   we would have a problem with synchronization in the smp case */
}

MODULE_AUTHOR("FSMLabs <support@fsmlabs.com>");
MODULE_DESCRIPTION("Real-Time Linux Main Module");
int quiet;
MODULE_PARM(quiet, "i");

int init_module(void)
{
	int ret;
	if ( arch_takeover() ) {
		printk("arch_takeover failed\n");
		return -1;
	}
	if ( !quiet ) {
		printk("Real-Time Linux Extensions Loaded (http://www.fsmlabs.com/)\n");
	}

	ret = rtl_printf_init();
	if (ret < 0) {
		return ret;
	}
	rtl_soft_sti();
	return 0;
}

void cleanup_module(void)
{
	HardDeclareAndInit(cpu_id);
	rtl_printf_cleanup();

	/*
	 * Process any pending interrupts, _hard_ disable
	 * then go on.  This way, we don't get any interrupts
	 * while we're vulnerable and giving up the architecture.
	 *   -- Cort
	 */
	rtl_hard_cli();
	rtl_soft_sti_no_emulation();
	do {
		rtl_hard_sti();
		rtl_process_pending();
		rtl_hard_cli();
	} while ( G_TEST(g_pend_since_sti) || L_TEST(l_pend_since_sti));
	
	arch_giveup();
}

spinlock_t debug_lock = SPIN_LOCK_UNLOCKED;

void rtl_debug(void)
{
	int i;
	unsigned long flags, xxx_last_cli = last_cli;

	rtl_spin_lock(&debug_lock);
	rtl_hard_save_flags(flags);
	printk( "RTL: cpu %d\n", smp_processor_id() );
	if ( ARCH_DEFINED_ENABLE == 0 )
		printk( "RTL: hard flags %08x %s\n", (unsigned int) flags,
			(flags&ARCH_DEFINED_DISABLE) ? "disabled" : "enabled" );
	else
		printk( "RTL: hard flags %08x %s\n", (unsigned int) flags,
			(flags&ARCH_DEFINED_ENABLE) ? "enabled" : "disabled" );
		
	printk( "RTL: global flags %lx %s%s\n", rtl_global.flags,
		((rtl_global.flags>>g_pend_since_sti)&1) ? "pend_since_sti " : "",
		((rtl_global.flags>>g_initializing)&1) ? "initializing " : "");
	printk( "RTL: global pending %lx %lx %lx %lx\n", rtl_global.pending[0],
		rtl_global.pending[1], rtl_global.pending[2],
		rtl_global.pending[3]);
	printk( "RTL: global enabled " );
	for ( i = 0; i < IRQ_ARRAY_SIZE; i++ )
		printk( "%lx ", rtl_global.soft_enabled[i] );
	printk( "\n" );
	for ( i = 0 ; i < NR_CPUS; i++ )
	{
		printk( "RTL: cpu%d "
#ifdef __LOCAL_IRQS__
			"local pending %x "
#endif			
			"flags: %x %s%s%s\n", i,
			rtl_local[i].flags,
#ifdef __LOCAL_IRQS__
			rtl_local[i].pending,
#endif			
			((rtl_local[i].flags>>l_ienable)&1)?"ienabled ":"disabled ",
			((rtl_local[i].flags>>l_pend_since_sti)&1)?"pend_since_sti ":"",
			((rtl_local[i].flags>>l_busy)&1)?"busy":"" );
	}
	printk( "RTL: last soft cli from: %08lx\n", xxx_last_cli );
	rtl_spin_unlock(&debug_lock);
}

void rtl_soft_irq_type(int unused, void *junk, struct pt_regs *garbage)
{
	printk("rtl_soft_irq_type(): shouldn't have been called!\n");
}

/* TODO VY: needs some synchronization here. Doesn't request_irq also
   have a problem? */
int rtl_get_soft_irq (void (*handler) (int, void *, struct pt_regs *),
                     const char *devname)
{
	int i;
	for (i = NR_IRQS - 1; i > 15; i--)
	{
		if (!linux_irq_desc[i].action)
		{
			int debug;
			if ((debug=request_irq (i, handler, 0, devname, 0)))
			{
				printk("RTL_GET_SOFT_IRQ %d: request=%d\n",
				       i, debug);
				return -EBUSY;
			}
			else
			{
				rtl_virt_enable(i);
				return i;
			}
		}
	}
	return -1;
}


