#!/bin/bash

BASIC="rtl rtl_time rtl_sched rtl_posixio rtl_fifo"

SETCOLOR_SUCCESS="echo -en \\033[1;32m"
SETCOLOR_FAILURE="echo -en \\033[1;31m"
SETCOLOR_WARNING="echo -en \\033[1;33m"
SETCOLOR_NORMAL="echo -en \\033[0;39m"

sync;sync;sync

cleanup()
{
    for x in periodic_test rtl_fifo rtl_posixio rtl_sched rtl_time rtl
    do
	(rmmod $x >& /dev/null)
    done
}

fatal()
{
    echo -n '[ '
    $SETCOLOR_FAILURE
    echo -n FAILED
    $SETCOLOR_NORMAL
    echo ' ]'
    echo 'Fatal Error.  Exiting.'
    cleanup
    exit 1
}

failure()
{
    echo -n '[ '
    $SETCOLOR_FAILURE
    echo -n FAILED
    $SETCOLOR_NORMAL
    echo ' ]'
}

success()
{
    echo -n '[ '
    $SETCOLOR_SUCCESS
    echo -n "  OK  "
    $SETCOLOR_NORMAL
    echo ' ]'
}

echo '----------------------------'
echo '- Installing basic modules -'
echo '----------------------------'
cleanup
for x in rtl rtl_time rtl_sched rtl_posixio rtl_fifo
do
    echo -n "Testing multiple loads of $x.o...	 	"
    l=`echo $x|wc -c`
    if [ `expr $l \< 10` = "1" ] ; then
    	echo -n "	"
    fi
    CNT=50
    while [ "${CNT}" != 0 ]; do
	CNT=`expr $CNT - 1`
	if [ "$x" = "rtl" ] ; then
		insmod modules/$x.o quiet=1 >& /dev/null
	else
		insmod modules/$x.o >& /dev/null
	fi
	if [ $? != "0" ] ; then
	    fatal
	    exit -1
	fi
	# on the last run, leave the module so others can load
	if [ "$CNT" != 0 ]; then
		rmmod $x >& /dev/null
	fi
    done
    success
done

sleep 1

echo -n 'Testing ping flood...					'
sync;sync;sync
(ping -f -c 100000 localhost >& /dev/null )
success # if it doesn't crash - it passes

echo -n 'Testing periodic timer...				'
(rmmod rtl_sched >& /dev/null)
(insmod examples/regression/periodic_test.o >& /dev/null)
if [ $? != "0" ] ; then
	failure
else
	(examples/regression/periodic_monitor >& /dev/null)
	if [ $? != "0" ] ; then
  		failure
	else
   		success
	fi
fi
(rmmod periodic_test >& /dev/null)

#  echo -n 'Testing Frank Zappa...	'
#  (insmod examples/frank/frank_module.o >& /dev/null)
#  if [ $? != "0" ] ; then
#      failure
#  else
#      (examples/frank/frank_app >& /dev/null)
#      if [ $? != "0" ] ; then
#  	failure
#      else
#     	success
#      fi
#  fi
#  (rmmod frank_module.o >& /dev/null)

cleanup

